/*
 * $Id: sufary.h,v 1.30 1999/12/05 03:22:32 tatuo-y Exp $
 *
 *  sufary.h --- SUFARY Library Header File
 *
 */

#ifndef _SUFARY_H_
#define _SUFARY_H_

#include <sys/types.h>
#include <stdarg.h>


#ifndef KEYWORD_MAX_LENGTH
#define KEYWORD_MAX_LENGTH 5000
#endif

#define SA_FILE_NAME_MAX 1000

typedef long SA_INDEX;


typedef enum sa_error_ {
    OTHER_ERROR = 0,
    MEMORY_ALLOCATE_ERROR = 1,
    FILE_OPEN_ERROR = 2,
    MMAP_ERROR = 3,
    KEYWORD_PARSING_ERROR = 4
} SA_ERROR;

extern int sa_error_no;
extern char *sa_error_str[];

#define sa_get_error_str ((char*)sa_error_str[sa_error_no])


typedef enum sa_stat_ {
    SUCCESS,
    FAIL
} SA_STAT;

typedef enum mmap_open_mode_ {
    SA_MMAP_RO,
    SA_MMAP_RW
} SA_MMAP_MODE;


typedef struct {
    off_t sz;
    caddr_t map;
    int is_fail;
} SA_MMAP;


typedef struct {
    SA_MMAP *mmtxt;
    SA_MMAP *mmary;
    char* text;		/* ƥȥեΥޥåץɥ쥹 */
    SA_INDEX* array;
    SA_INDEX textsize;		/* ƥȥեΥ */
    SA_INDEX arraysize;		/* Array 礭 */
} SUFARY;


typedef struct {
    SUFARY *suf;
    SA_INDEX left;			/* ϰϤκü(ϰϤ¦ؤ) */
    SA_INDEX right;			/* ϰϤαü(ϰϤ¦ؤ) */
    SA_STAT stat;
} SUF_RESULT;


typedef struct {
    SA_MMAP *mmdid;
    SA_INDEX *docid;		/* mmap줿 */
    SA_INDEX didsize;		/* DID 礭 */
} DID;


typedef struct {
    SA_INDEX start;			/* ɥȳϰ */
    SA_INDEX size;			/* ɥȥ */
    SA_INDEX no;			/* ɥֹ */
    SA_STAT stat;
} DID_RESULT;


typedef enum sa_mki_mode_ {
    SA_AFTER = 1,
    SA_AT = 2,
    SA_EUC_2 = 4,
    SA_CUT_TOP = 8,
    SA_ONLY_00_7F = 16,
    SA_ONLY_80_FF = 32,
    SA_VERBOSE = 64
} SA_MAKE_INDEX_MODE;


/* ̥ꥹ(for regex) */
typedef struct REGEX_RESULT_LIST REGEX_RESULT_LIST;
struct REGEX_RESULT_LIST {
    SA_INDEX value;
    int len;
    REGEX_RESULT_LIST *next;
};


/* in search.c */
extern SUF_RESULT sa_find(SUFARY * ary, SA_INDEX left, SA_INDEX right,
			  char *keyword, int keyword_length,
			  int base_txt_skip);
extern char *sa_aryidx2txtptr(SUFARY * ary, SA_INDEX pos);
extern SA_INDEX sa_aryidx2txtidx(SUFARY * ary, SA_INDEX idx);
extern char *sa_txtidx2txtptr(SUFARY * ary, SA_INDEX idx);

extern DID_RESULT sa_didsearch(DID * did, SA_INDEX target);
extern SA_INDEX sa_get_start_position(DID * did, SA_INDEX id);
extern SA_INDEX sa_get_end_position(DID * did, SA_INDEX id);
extern SA_INDEX sa_dididx2txtidx(DID * did, SA_INDEX idx);

extern char *sa_get_block(SUFARY * ary, SA_INDEX pos, char *d1, char *d2);
extern char *sa_get_line(SUFARY * ary, SA_INDEX pos);
extern char *sa_get_lines(SUFARY * ary, SA_INDEX pos, int B, int A);
extern char *sa_get_string(SUFARY * ary, SA_INDEX from, SA_INDEX size);

extern SA_INDEX sa_get_position_in_string(void);


/* in regex.c */
extern REGEX_RESULT_LIST *sa_regex(SUFARY * ary, SA_INDEX left, SA_INDEX right,
				   char *keyword, int keyword_length);
extern void free_regex_result_list(REGEX_RESULT_LIST * listp);

extern void sa_set_debug_mode(int i);


/* in file.c */
extern SUFARY *sa_open(char *text_file_name, char *array_file_name);
extern void sa_close(SUFARY * ary);

extern char *sa_get_text_ptr(SUFARY * ary);
extern SA_INDEX sa_get_text_size(SUFARY * ary);
extern SA_INDEX* sa_get_array_ptr(SUFARY * ary);
extern SA_INDEX sa_get_array_size(SUFARY * ary);

extern DID *sa_open_did(char *fn);
extern void sa_close_did(DID * d);

extern SA_INDEX *sa_get_did_ptr(DID * did);
extern SA_INDEX sa_get_did_size(DID * did);

extern size_t sa_fwrite(SA_INDEX idx, FILE *fd);

extern SA_MMAP *sa_open_mmap(char *fname, SA_MMAP_MODE mode);
extern void sa_close_mmap(SA_MMAP *mmr);

extern off_t sa_get_mmap_size(SA_MMAP *mmr);
extern caddr_t sa_get_mmap_ptr(SA_MMAP *mmr);


/* in make-index.c */
extern SA_STAT sa_sort_index(char* txt_fname, char* ary_fname);
extern SA_STAT sa_sort_index_block(char* txt_fname, char* ary_fname,
				    int num_of_block);

extern SA_STAT sa_write_index(char *in_fname, char *ary_fname, int mode);

extern SA_STAT sa_make_did(char *txt_fname, char *ary_fname, char* did_fname,
			    char* tag1, char* tag2, int mki_mode);

extern void sa_set_ignored_chars(char* p);
extern char *sa_get_ignored_chars(void);
extern void sa_set_my_chars(char* p);
extern char *sa_get_my_chars(void);
extern void sa_set_my_delimiters(char* p);
extern char *sa_get_my_delimiters(void);


#endif				/* _SUFARY_H_ */
