/*
 * $Id: sass.c,v 1.25 1999/12/05 07:18:54 tatuo-y Exp $
 *
 * sass --- Suffix Array Simple Search
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "sufary.h"
#include "getopt.h"

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#define STR_LEN_MAX 1000

int regex_mode = 0;
int stdin_mode = 0;
int ignore_case_mode = 0;
int before_context = 0;
int after_context = 0;
int keyword_emphasis_mode = 0;
int tag_mode = 0;
char tag_str[STR_LEN_MAX];

static void search_and_print(SUFARY *ary, char *key, char *fname);
static void show_usage(void);
static void show_mini_usage(void);
static int parse_options(int, char**);
static int ck_atoi(char const *, int *);
/*
 * Command line options.
 */
static const char *short_options = "0123456789hirsA:B:C::123456789kvDt:";
static struct option long_options[] = {
    {"help",             no_argument,       NULL, 'h'},
    {"ignore-case",      no_argument,       NULL, 'i'},
    {"regex",            no_argument,       NULL, 'r'},
    {"stdin",            no_argument,       NULL, 's'},
    {"keyword-emphasis", no_argument,       NULL, 'k'},
    {"after-context",    required_argument, NULL, 'A'},
    {"before-context",   required_argument, NULL, 'B'},
    {"context",          optional_argument, NULL, 'C'},
    {"tags",             optional_argument, NULL, 't'},
    {"debug",            no_argument,       NULL, 'D'},
    {"version",          no_argument,       NULL, 'v'},
    {"1",                no_argument,       NULL, '1'},
    {"2",                no_argument,       NULL, '2'},
    {"3",                no_argument,       NULL, '3'},
    {"4",                no_argument,       NULL, '4'},
    {"5",                no_argument,       NULL, '5'},
    {"6",                no_argument,       NULL, '6'},
    {"7",                no_argument,       NULL, '7'},
    {"8",                no_argument,       NULL, '8'},
    {"9",                no_argument,       NULL, '9'},
    {NULL, 0, NULL, 0}
};


int main(int argc, char *argv[])
{
    SUFARY *ary[256];
    int i, argc_offset;
    int num_of_files;
    char* key = NULL;

    argc_offset = parse_options(argc, argv);

    if (argc_offset == argc) {
	show_mini_usage();
	exit(1);
    }

    if (!stdin_mode) {
	key = argv[argc_offset];
	argc_offset++;
    }

    num_of_files = argc - argc_offset;
    if (num_of_files < 1) {
	show_mini_usage();
	exit(1);
    }

    for (i = 0; i < num_of_files; i++) {
	if ((ary[i] = sa_open(argv[i + argc_offset], NULL)) == NULL) {
	    exit(1);
	}
    }

    if (stdin_mode) {
	/* keyword from stdin */
	char cmd[1000];
	while (fgets(cmd, (int)sizeof(cmd), stdin)) {
	    cmd[strlen(cmd) - 1] = '\0'; /* delete '\n' */
	    for (i = 0; i < num_of_files; i++) {
		if (num_of_files == 1)
		    search_and_print(ary[i], cmd, NULL);
		else
		    search_and_print(ary[i], cmd, argv[i + 1]);
	    }
	}
    } else {
	/* keyword from argv */
	for (i = 0; i < num_of_files; i++) {
	    if (num_of_files == 1)
		search_and_print(ary[i], key, NULL);
	    else
		search_and_print(ary[i], key, argv[i + 1]);
	}
    }

    for (i = 0; i < num_of_files; i++)
	sa_close(ary[i]);

    return 0;
}


/*
 */
void search_and_print_sub(SUFARY *ary, long pos, char *fname, int key_len)
{
    char* s;
    int kpos_in_str;

    if (tag_mode) {
	int i;
	char *tag1_ptr = tag_str;
	char *tag2_ptr;
	for (i = 0; i < strlen(tag_str); i++)
	    if (tag_str[i] == ',') {
		tag_str[i] = '\0';
		tag2_ptr = tag_str + i + 1;
		break;
	    }
	s = sa_get_block(ary, pos, tag1_ptr, tag2_ptr);
    } else {
	s = sa_get_lines(ary, pos, before_context, after_context);
    }

    kpos_in_str = sa_get_position_in_string();

    if (fname != NULL)
	printf("%s:", fname);

    if (keyword_emphasis_mode) {

	printf("%.*s", kpos_in_str, s);
	printf("\x1b[7m");
/*	
	int i;
	for (i = 0; i < key_len; i++) {
	    char tmp = *(s + kpos_in_str + i);
	    printf("%c\x08%c\x08%c\x08%c", tmp, tmp, tmp, tmp);
	}
	*/
	printf("%.*s", key_len, s + kpos_in_str);
	printf("\x1b[0m");

	printf("%.*s\n", (int)strlen(s) - kpos_in_str - key_len,
	       s + kpos_in_str + key_len);
    } else {
	printf("%s\n", s);
    }
    if (after_context >= 1 || before_context >= 1 || tag_mode) {
	printf("--\n");
    }

    free(s);
}


/*
 * find a keyword and print chunks
 */
void search_and_print(SUFARY *ary, char *key, char *fname)
{
    if (strlen(key) == 0)
	return;

    if (regex_mode || ignore_case_mode) {
	REGEX_RESULT_LIST *ll, *tmp;
	char *key_tmp = key;

	if (ignore_case_mode) {
	    /* change key to regex string. sample: the -> [tT][hH][eE] */
	    int i, j = 0;
	    key_tmp = malloc(strlen(key) * 4 + 1);
	    for (i = 0; i < strlen(key); i++) {
		if (isalpha(key[i])) {
		    key_tmp[j] = '[';
		    key_tmp[j + 1] = tolower(key[i]);
		    key_tmp[j + 2] = toupper(key[i]);
		    key_tmp[j + 3] = ']';
		    j += 4;
		} else if (key[i] == '[' || key[i] == ']' || key[i] == '\\' ||
			   key[i] == '(' || key[i] == ')' || key[i] == '.' ||
			   key[i] == '|' || key[i] == '*' || key[i] == '?') {
		    key_tmp[j] = '\\';
		    key_tmp[j + 1] = key[i];
		    j += 2;
		} else {
		    key_tmp[j] = key[i];
		    j++;
		}
	    }
	    key_tmp[j] = '\0';
	}

        tmp = sa_regex(ary, 0, ary->arraysize - 1, key_tmp, strlen(key_tmp));

        for (ll = tmp; ll != NULL; ll = ll->next) {
	    long pos = ll->value;
	    search_and_print_sub(ary, pos, fname, ll->len);
        }

        free_regex_result_list(tmp);

    } else {
	SUF_RESULT sr;
	sr = sa_find(ary, 0, ary->arraysize - 1, key, strlen(key), 0);

	if (sr.stat == SUCCESS) {
	    SA_INDEX ai;
	    for (ai = sr.left; ai <= sr.right; ai++) {
		SA_INDEX pos = sa_aryidx2txtidx(ary, ai);
		search_and_print_sub(ary, pos, fname, strlen(key));
	    }
	}

    }

}


void show_usage(void)
{
    printf("\
sass %s, simple search program of SUFARY.\n\n\
Usage: sass [options] <pattern> <file>...\n\
  -r, --regex               use regex for the pattern.\n\
  -i, --ignore-case         ignore case.\n\
  -s, --stdin               read the pattern from stdin.\n\
  -k, --keyword-emphasis    emphasize keyword in results\n\
  -A, --after-context=NUM   print NUM lines of trailing context\n\
  -B, --before-context=NUM  print NUM lines of leading context\n\
  -C, --context[=NUM]       print NUM (default 2) lines of output context\n\
                            unless overriden by -A or -B\n\
  -NUM                      same as --context=NUM\n\
  -t, --tags=TAG1,TAG2      print text regions (from TAG1 to TAG2).\n\
                            example: -t '<doc>,</doc>', -t '\\n\\n,\\n\\n'\n\
  -v, --version             show the version of SUFARY and exit.\n\
  -D, --debug               show the debug information.\n\n\
Notes:\n\
  You should prepare an array file for each target file.\n\
  (these file names must be its original file name followed by \".ary\")\n",
	   VERSION);
}

void show_mini_usage(void)
{
    printf("\
Usage: sass [options] <pattern> <file ...>\n\
Try `sass --help' for more information.\n");
}

int parse_options(int argc, char **argv)
{
    int digit_args_val = 0;

    while (1) {
        int ch = getopt_long(argc, argv, short_options, long_options, NULL);
        if (ch == EOF) {
            break;
	}
	switch (ch) {
	case 'h':
	    show_usage();
	    exit(1);
	    break;
	case 'i':
	    ignore_case_mode = 1;
	    break;
	case 'r':
	    regex_mode = 1;
	    break;
	case 's':
	    stdin_mode = 1;
	    break;
	case 'A':
	    if (ck_atoi(optarg, &after_context)) {
		fprintf(stderr, "invalid context length argument\n");
		exit(1);
	    }
	    break;
	case 'B':
	    if (ck_atoi(optarg, &before_context)) {
		fprintf(stderr, "invalid context length argument\n");
		exit(1);
	    }
	    break;
	case 'C':
	    if (optarg) {
		if (ck_atoi(optarg, &before_context)) {
		    fprintf(stderr, "invalid context length argument\n");
		    exit(1);
		}
		after_context  = before_context;
	    } else {
		before_context = 2;
		after_context  = 2;
	    }

	    break;
	case 'D':
	    sa_set_debug_mode(1);
	    break;
	case 't':
	    tag_mode = 1;
	    {
		int i;
		int j = 0;
		for (i = 0; i <= strlen(optarg); i++) {
		    if (optarg[i] == '\\' && i < strlen(optarg) &&
                        optarg[i + 1] == 'n') {
			tag_str[j] = '\n';
			i++;
		    } else {
			tag_str[j] = optarg[i];
		    }
		    j++;
		}
	    }
	    break;
	case 'k':
	    keyword_emphasis_mode = 1;
	    break;
	case 'v':
            printf("sass of SUFARY %s\n", VERSION);
	    printf("Copyright (C) 1997-1999 YAMASHITA Tatuo All rights reserved.\n");
            exit(1);
	    break;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    digit_args_val = 10 * digit_args_val + ch - '0';
	    before_context = digit_args_val;
	    after_context  = digit_args_val;
	    break;
	default:
	    show_mini_usage();
	    exit(1);
	}
    }
    return optind;
}

/* Imported from GNU grep-2.3 [1999-11-08] by satoru-t */

/* Convert STR to a positive integer, storing the result in *OUT.
   If STR is not a valid integer, return -1 (otherwise 0). */
static int
ck_atoi (str, out)
     char const *str;
     int *out;
{
  char const *p;
  for (p = str; *p; p++)
    if (*p < '0' || *p > '9')
      return -1;

  *out = atoi (optarg);
  return 0;
}
